"""
BlackDog

Copyright (C) 2014 Snaipe, Therozin

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import logging
import re

from pyquery import PyQuery
import requests

from blackdog.plugin import Plugin, PluginStage, PluginVersion


class BukkitDev(object):

    def __init__(self, cache_dir):
        self.cache_dir = cache_dir
        self.base = 'http://dev.bukkit.org'
        self.logger = logging.getLogger('BukkitDev')

    def save_plugin(self, plugin: Plugin):
        return plugin.save(self.cache_dir)

    def load_plugin(self, plugin: Plugin):
        return plugin.load(self.cache_dir)

    def _fill_version_meta(self, version: PluginVersion, metalink):
        d = PyQuery(url=self.base + metalink)
        path = '#bd section.main .main-body .main-body-inner .line .lastUnit .content-box .content-box-inner dl'
        info = d(path).children()
        meta = dict([(PyQuery(dt).text(), PyQuery(dd)) for dt, dd in zip(info[::2], info[1::2])])

        version.md5(meta['MD5'].text())
        version.stage(meta['Type'].text())
        version.date(meta['Uploaded on'].text())
        version.game_versions([v.text() for v in meta['Game version']('ul li').items()])
        version.url(meta['Filename']('a').eq(0).attr('href'))

    def _fill_plugin_meta(self, plugin: Plugin, version=None):
        try:
            if not requests.head('/'.join([self.base, 'bukkit-plugins', plugin.name, ''])).ok:
                plugin.exists(False)
                return

            page = 1
            while True:
                page_url = '/'.join([self.base,
                                     'bukkit-plugins',
                                     plugin.name,
                                     'files',
                                     '?page=%s' % page if page > 1 else ''])

                code = requests.head(page_url).status_code
                if code != 200:
                    break

                d = PyQuery(url=page_url)
                list = [tr for tr in d('table.listing tbody tr').items()
                        if tr('td.col-filename').text().endswith('.jar')]

                for tr in list:
                    try:
                        version_str = re.sub(r'.*?([0-9]+(\.[0-9]+)+).*', '\\1', tr('td.col-file').text())
                        metalink = tr('td.col-file a').attr('href')
                        version_found = version == version_str or version == 'latest'
                        if metalink and version_str and (not version or version_found):
                            pversion = PluginVersion(plugin, version_str)
                            self._fill_version_meta(pversion, metalink)
                            plugin.add_version(pversion)
                            if version_found:
                                break
                    except:
                        self.logger.error('Could not retrieve metadata for plugin %s', plugin.name)
                        continue
                else:
                    page += 1
                    continue
                break
        finally:
            self.save_plugin(plugin)

    @staticmethod
    def _to_post_arg(args):
        """
        Transforms a dictionary to POST parameters:
        {a: 1, b: '2', c='a b'} -> 'a=1&b=2&c=a+b'
        :param args: the dictionary to transform
        :return: the POST parameters
        """
        arg_list = []
        for key, value in args.items():
            arg_list.append(key + '=' + re.sub(r'[ \t\r\n]+', '+', str(value)))
        return '&'.join(arg_list)

    def get_plugin(self, name, version=None, no_query=False):
        """
        Gets a plugin info from its name and version
        :param name: the plugin name
        :param version: the plugin version
        :return: the plugin
        """

        plugin = Plugin(name).load(self.cache_dir)

        if not no_query:
            self._fill_plugin_meta(plugin, version)

        return plugin

    def search(self, **kwargs):
        """
        Searchs for a plugin
        :param search: the query string
        :param page: the result page number, first page by default
        :param category: the plugin category, or all categories if None
        :param stage: the development stage
        :return: a list of matching plugins
        """

        results = []
        page_content = PyQuery(url='/'.join([self.base, 'bukkit-plugins', '?%s' % self._to_post_arg(kwargs)]))
        plugin_table = page_content('#bd .line .unit .listing-container .listing-container-inner table tbody tr')

        for (info, summary) in zip(plugin_table[::2], plugin_table[1::2]):
            info = PyQuery(info)
            plugin = Plugin(name=info('td.col-project h2 a').attr('href').split('/')[2])
            plugin.display_name(info('td.col-project').text())
            plugin.stage(PluginStage.from_string(info('td.col-status').text()))
            plugin.summary(PyQuery(summary)('td.summary').text())
            results.append(plugin)

        return results

    def scan(self, stages=None):
        if not stages:
            stages = [PluginStage.release, PluginStage.mature]

        self.logger.info('Scanning http://dev.bukkit.org/...')
        for stage in stages:
            self.logger.info('Processing plugins for stage \'%s\'', stage.name)
            page = 1
            more = True
            while more:
                self.logger.info('processing page %s...', page)
                plugins = self.search(stage=stage.value, page=page)
                for plugin in plugins:
                    try:
                        self._fill_plugin_meta(plugin)
                    except Exception:
                        self.logger.error('Could not process plugin %s', plugin.name)

                more = len(plugins) > 0
                page += 1


"""
BlackDog

Copyright (C) 2014 Snaipe, Therozin

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import re
from configparser import ConfigParser
from enum import Enum
from functools import wraps


def config_node(node, type=str, section=None):
    def decorator(fun):
        config_field = '__config__' + re.sub(r'[^a-zA-Z0-9_]', '_', node)

        @wraps(fun)
        def wrapper(self, *args, **kwargs):
            if len(args) == 1:
                setattr(self, config_field, args[0])
            elif 'value' in kwargs:
                setattr(self, config_field, kwargs['value'])
            else:
                if not hasattr(self, config_field) or 'default' in kwargs and kwargs['default']:
                    ret = fun(self)
                    return ret

                return getattr(self, config_field)

        wrapper.config_node = node
        wrapper.config_type = type
        wrapper.config_section = section

        return wrapper
    return decorator


def load(config: ConfigParser, obj, section):
    attrs = [(f, getattr(obj, f)) for f in dir(obj)]
    for name, method in [(n, m) for (n, m) in attrs if callable(m) and hasattr(m, 'config_node')]:
        real_section = method.config_section if method.config_section else section
        val = config.get(real_section, method.config_node, fallback=None)

        if val:
            if method.config_type == list:
                val = [e.strip() for e in val.split(',')]
            elif method.config_type == bool:
                try:
                    val = config.getboolean(real_section, method.config_node, fallback=None)
                except ValueError:
                    val = None
            elif issubclass(method.config_type, Enum):
                val = method.config_type.from_string(val)

            method(val)
        else:
            method(method(default=True))


def save(config: ConfigParser, obj, section):
    attrs = [(f, getattr(obj, f)) for f in dir(obj)]
    nodes = [(n, m) for (n, m) in attrs if callable(m) and hasattr(m, 'config_node')]
    for name, method in nodes:
        real_section = method.config_section if method.config_section else section
        val = method()

        if val:
            if method.config_type == list:
                val = ', '.join(val)
            elif isinstance(val, Enum):
                val = val.name
            else:
                val = str(val)

            if not config.has_section(real_section):
                config.add_section(real_section)
            config.set(real_section, method.config_node, val)


"""
BlackDog

Copyright (C) 2014 Snaipe, Therozin

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import hashlib
import re
from http.server import SimpleHTTPRequestHandler
from socketserver import TCPServer
import atexit

import requests

from blackdog import PluginVersion, BlackDogException


class HTTPServer(TCPServer):

    def __init__(self, port):
        super().__init__(("", port), RequestHandler, bind_and_activate=False)
        from blackdog import BlackDog
        self.blackdog = BlackDog.instance
        self.port = port

    def __enter__(self):
        self.server_bind()
        self.server_activate()
        atexit.register(self.close)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        atexit.unregister(self.close)

    def close(self):
        self.server_close()


def pattern(pattern):
    def decorator(func):
        func.pattern = pattern
        return func
    return decorator


class RequestHandler(SimpleHTTPRequestHandler):

    def get_groupid(self):
        return '.'.join(self.path.split('/')[1:-3])

    def handle_pattern(self, path, *args, **kwargs):
        handles = []
        for m in [m for m in [getattr(self, m) for m in dir(self)] if hasattr(m, 'pattern')]:
            handles.append((m.pattern, m))

        for p, m in handles:
            if re.match(p, path):
                m(*args, **kwargs)
                return True
        return False

    def handle_text(self, content, mime='text/plain'):
        if not content:
            self.send_response(404)
            self.end_headers()
            return

        self.send_response(200)
        self.send_header('Content-type', mime)
        self.send_header('Content-length', len(content))
        self.end_headers()
        self.wfile.write(content.encode('ascii'))

    @pattern(r'.*\.jar$')
    def handle_jar(self, version: PluginVersion):
        if not version.url():
            self.send_response(404)
            self.end_headers()
            return

        r = requests.get(version.url(), stream=True)
        if r.status_code != 200:
            self.send_response(404)
            self.end_headers()
            r.close()
            return

        self.send_response(200)
        self.send_header('Content-type', 'application/java-archive')
        self.send_header('Content-length', r.headers['content-length'])
        self.end_headers()

        if not version.sha1():
            sha1 = hashlib.sha1()
            for chunk in r.iter_content(16 * 1024):
                self.wfile.write(chunk)
                sha1.update(chunk)
            version.sha1(sha1.hexdigest())
            self.server.blackdog.bukkitdev.save_plugin(version.get_plugin())
        else:
            for chunk in r.iter_content(16 * 1024):
                self.wfile.write(chunk)

    @pattern(r'.*\.pom$')
    def handle_pom(self, version: PluginVersion):
        self.handle_text(version.get_pom(self.get_groupid()), mime='text/xml')

    @pattern(r'.*\.jar.sha1$')
    def handle_jar_sha1(self, version: PluginVersion):
        self.handle_text(version.sha1())

    @pattern(r'.*\.jar.md5$')
    def handle_jar_md5(self, version: PluginVersion):
        self.handle_text(version.md5())

    @pattern(r'.*\.pom.sha1$')
    def handle_pom_sha1(self, version: PluginVersion):
        self.handle_text(version.get_pom_sha1(self.get_groupid()))

    @pattern(r'.*\.pom.md5$')
    def handle_pom_md5(self, version: PluginVersion):
        self.handle_text(version.get_pom_md5(self.get_groupid()))

    def do_GET(self):
        from blackdog import BlackDog
        blackdog = BlackDog.instance

        try:
            dir_path = self.path.split('/')[:-1]
            plugin_version = dir_path[-1]
            plugin_name = dir_path[-2]
        except:
            self.send_response(404)
            self.end_headers()
            return

        try:
            plugin = blackdog.bukkitdev.get_plugin(plugin_name, no_query=True)
            if not plugin.has_version(plugin_version) or not plugin.get_version(plugin_version).can_download():
                plugin = blackdog.bukkitdev.get_plugin(plugin_name, version=plugin_version)

            version = plugin.get_version(plugin_version)

            if not self.handle_pattern(self.path, version):
                self.send_response(404)
                self.end_headers()

        except BlackDogException as e:
            blackdog.logger.error(e.message)
            self.send_response(404)
            self.end_headers()
        except Exception as e:
            blackdog.logger.exception(e)
            self.send_response(404)
            self.end_headers()


"""
BlackDog

Copyright (C) 2014 Snaipe, Therozin

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import errno
import logging
import os
from os.path import exists, expanduser, join

from baker import run

from blackdog.exception import *
from blackdog.bukkitdev import BukkitDev, PluginStage, Plugin, PluginVersion
from blackdog.server import HTTPServer


class BlackDog(object):
    instance = None

    def __init__(self):
        BlackDog.instance = self

        logging.basicConfig(level='INFO', format="[%(name)s] %(message)s")

        self.logger = logging.getLogger('BlackDog')
        self.directory = expanduser('~/.blackdog/')
        os.makedirs(self.directory, mode=0o755, exist_ok=True)
        self.bukkitdev = BukkitDev(self.directory)
        self.pidfile = join(self.directory, '.pid')

    def main(self):
        try:
            run()
        except BlackDogException as ex:
            self.logger.error(ex.message)

    @staticmethod
    def checkpid(pid):
        try:
            os.kill(pid, 0)
        except OSError as e:
            if e.errno == errno.ESRCH:
                return False
        return True

    def get_server_pid(self):
        if exists(self.pidfile):
            with open(self.pidfile, 'r') as f:
                return int(f.read())

    def is_server_running(self):
        pid = self.get_server_pid()
        return self.checkpid(pid) if pid else False


"""
BlackDog

Copyright (C) 2014 Snaipe, Ojukashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from string import Template


class BlackDogException(Exception):
    def __init__(self, message, *args, **kargs):
        super().__init__(*args)
        self.message = Template(message).substitute(**kargs)


class ServerAlreadyRunningException(BlackDogException):
    def __init__(self):
        super().__init__('Server is already running')


class ServerNotRunningException(BlackDogException):
    def __init__(self):
        super().__init__('Server is not running')


class NoSuchPluginException(BlackDogException):
    def __init__(self, plugin):
        super().__init__('Plugin ${name} could not be found', plugin, name=plugin.name)


class NoSuchPluginVersionException(BlackDogException):
    def __init__(self, version):
        super().__init__('Version ${version} could not be found', version, version=version)


"""
BlackDog

Copyright (C) 2014 Snaipe, Therozin

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import hashlib
import re
from configparser import ConfigParser
from enum import Enum
from string import Template

from blackdog import NoSuchPluginVersionException
from blackdog.config import load, save, config_node


class PluginStage(Enum):
    planning = 'p'
    alpha = 'a'
    beta = 'b'
    release = 'r'
    mature = 'm'
    inactive = 'i'
    abandoned = 'x'
    deleted = 'd'

    @classmethod
    def from_string(cls, string):
        return getattr(PluginStage, string.lower(), None) if string else None


class Plugin(object):
    def __init__(self, name):
        self.name = name
        self.path_name = re.sub(r'[^a-zA-Z0-9_\-]', '_', name)
        self.versions = {}

    @config_node('summary')
    def summary(self):
        return None

    @config_node('display-name')
    def display_name(self):
        return self.name

    @config_node('stage')
    def stage(self):
        return None

    @config_node('exists', type=bool)
    def exists(self):
        return None

    def add_version(self, version):
        self.versions[version.get_version()] = version

    def get_version(self, version):
        try:
            return self.versions[version]
        except KeyError as e:
            raise NoSuchPluginVersionException(e.args[0])

    def has_version(self, version):
        return version in self.versions

    def _get_config(self, directory):
        from os.path import join
        return join(directory, self.path_name + '.data')

    def load(self, directory):
        """
        Loads all available plugin metadata from cache
        :param directory: the cache root directory
        :return: the plugin itself
        """
        config = ConfigParser()
        config.read(self._get_config(directory))

        load(config, self, 'plugin')
        for section in [s for s in config.sections() if s != 'plugin']:
            version = PluginVersion(self, section)
            load(config, version, section)
            self.versions[section] = version

        return self

    def save(self, directory):
        """
        Saves all available plugin metadata to cache
        :param directory: the cache root directory
        :return: the plugin itself
        """
        config = ConfigParser()

        save(config, self, 'plugin')
        for vstr, version in self.versions.items():
            save(config, version, vstr)

        with open(self._get_config(directory), 'w') as fd:
            config.write(fd)

        return self


class PluginVersion(object):
    __POM_BASE = \
        '<?xml version="1.0" encoding="UTF-8"?>\n' \
        '<project xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd"\n' \
        '    xmlns="http://maven.apache.org/POM/4.0.0"\n' \
        '    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">\n' \
        '    <modelVersion>4.0.0</modelVersion>\n' \
        '    <groupId>${groupid}</groupId>\n' \
        '    <artifactId>${artifactid}</artifactId>\n' \
        '    <version>${version}</version>\n' \
        '</project>'

    def __init__(self, plugin: Plugin, version):
        if not version or not plugin:
            raise ValueError('plugin or version cannot be None')
        self.__plugin = plugin
        self.__version = version

    @config_node('url')
    def url(self):
        return None

    @config_node('sha1')
    def sha1(self):
        return None

    @config_node('md5')
    def md5(self):
        return None

    @config_node('date')
    def date(self):
        return None

    @config_node('stage', type=PluginStage)
    def stage(self):
        return None

    @config_node('game-versions', type=list)
    def game_versions(self):
        return None

    def can_download(self):
        return self.url() and self.md5()

    def get_version(self):
        return self.__version

    def get_plugin(self):
        return self.__plugin

    def get_pom(self, groupid):
        return Template(PluginVersion.__POM_BASE).substitute(
            groupid=groupid,
            artifactid=self.__plugin.name,
            version=self.__version
        )

    def __get_pom_hash(self, groupid, hash):
        hash.update(self.get_pom(groupid).encode('ascii'))
        return hash.hexdigest()

    def get_pom_md5(self, groupid):
        return self.__get_pom_hash(groupid, hashlib.md5())

    def get_pom_sha1(self, groupid):
        return self.__get_pom_hash(groupid, hashlib.sha1())